/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LoHiSide.H"

#include "GaussianIBC.H"
#include "GaussianF_F.H"

#include "LoHiCenter.H"

// Null constructor
GaussianIBC::GaussianIBC()
{
    m_isFortranCommonSet = false;
}

// Constructor which defines parameters used by Fortran routines
GaussianIBC::GaussianIBC(const Real&     a_gamma,
                         const Real&     a_ambientDensity,
                         const Real&     a_deltaDensity,
                         const int&      a_pressure,
                         const RealVect& a_center,
                         const Real&     a_size,
                         const RealVect& a_velocity,
                         const Real&     a_artvisc)
{
    setFortranCommon(a_gamma,
                     a_ambientDensity,
                     a_deltaDensity,
                     a_pressure,
                     a_center,
                     a_size,
                     a_velocity,
                     a_artvisc);
}

// Sets parameters in a common block used by Fortran routines:
//   a_gamma          - Gamma for polytropic, gamma-law gas
//   a_ambientDensity - Ambient density add to the density gaussian
//   a_deltaDensity   - Mean of the gaussian
//   a_pressure       - If 0, use isentropic pressure
//                      if 1, use the constant pressure of 1.0
//   a_center         - Center of the gaussian
//   a_size           - Standard deviation of the gaussian
//   a_velocity       - Initial velocity of the gas
//   a_artvisc        - Artificial viscosity coefficient
void GaussianIBC::setFortranCommon(const Real&     a_gamma,
                                   const Real&     a_ambientDensity,
                                   const Real&     a_deltaDensity,
                                   const int&      a_pressure,
                                   const RealVect& a_center,
                                   const Real&     a_size,
                                   const RealVect& a_velocity,
                                   const Real&     a_artvisc)
{
    assert(m_isFortranCommonSet == false);

    FORT_SETGAUSSIANF(CHF_CONST_REAL(a_gamma),
                      CHF_CONST_REAL(a_ambientDensity),
                      CHF_CONST_REAL(a_deltaDensity),
                      CHF_CONST_INT(a_pressure),
                      CHF_CONST_REALVECT(a_center),
                      CHF_CONST_REAL(a_size),
                      CHF_CONST_REALVECT(a_velocity),
                      CHF_CONST_REAL(a_artvisc));

    m_isFortranCommonSet = true;
}

// Set the flag m_isFortranCommonSet to true so that new IBCs made with
// new_physIBC() will have this flag set without calling setFortranCommon()
// (this is a clumsy design and should be improved).
void GaussianIBC::setFortranCommonSet()
{
  m_isFortranCommonSet = true;
}

// Factory method - this object is its own factory:
//   Return a pointer to a new PhysIBC object with m_isDefined = false (i.e.,
//   its define() must be called before it is used) and m_isFortranCommonSet
//   set to value of m_isFortranCommonset in the current (factory) object.
PhysIBC* GaussianIBC::new_physIBC()
{
  GaussianIBC* retval = new GaussianIBC();

  if (m_isFortranCommonSet == true)
  {
    retval->setFortranCommonSet();
  }

  return static_cast<PhysIBC*>(retval);
}

// Set boundary fluxes
void GaussianIBC::fluxBC(FArrayBox&            a_F,
                         const FArrayBox&      a_W,
                         const FArrayBox&      a_Wextrap,
                         const int&            a_dir,
                         const Side::LoHiSide& a_side,
                         const Real&           a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  // In periodic case, this doesn't do anything
  if (!m_domain.isPeriodic(a_dir))
  {
    int sign;
    Box FBox = a_F.box();
    Box tmp = FBox;

    // Determine which side and thus shifting directions
    if (a_side == Side::Lo)
    {
      sign = -1;
    }
    else
    {
      sign = 1;
    }

    tmp.shiftHalf(a_dir,sign);

    // Is there a domain boundary next to this grid
    if (!m_domain.contains(tmp))
    {
      tmp &= m_domain;

      Box boundaryBox;

      // Find the strip of cells next to the domain boundary
      if (a_side == Side::Lo)
      {
        boundaryBox = bdryLo(tmp,a_dir);
      }
      else
      {
        boundaryBox = bdryHi(tmp,a_dir);
      }

      // Shift things to all line up correctly
      boundaryBox.shiftHalf(a_dir,-sign);
      a_F.shiftHalf(a_dir,-sign);

      // Set the boundary fluxes
      FORT_SOLIDBCF(CHF_FRA(a_F),
                    CHF_CONST_FRA(a_Wextrap),
                    CHF_CONST_INT(sign),
                    CHF_CONST_REAL(m_dx),
                    CHF_CONST_INT(a_dir),
                    CHF_BOX(boundaryBox));

      // Shift returned fluxes to be face centered
      a_F.shiftHalf(a_dir,sign);
    }
  }
}

// Set boundary slopes:
//   The boundary slopes in a_dW are already set to one sided difference
//   approximations.  If this function doesn't change them they will be
//   used for the slopes at the boundaries.
void GaussianIBC::setBdrySlopes(FArrayBox&       a_dW,
                                const FArrayBox& a_W,
                                const int&       a_dir,
                                const Real&      a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  // In periodic case, this doesn't do anything
  if (!m_domain.isPeriodic(a_dir))
    {
      Box loBox,hiBox,centerBox,domain;
      int hasLo,hasHi;
      Box slopeBox = a_dW.box()&m_domain;

      // Generate the domain boundary boxes, loBox and hiBox, if there are
      // domain boundarys there
      loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,domain,
                 slopeBox,m_domain,a_dir);

      // Set the boundary slopes if necessary
      if ((hasLo != 0) || (hasHi != 0))
        {
          FORT_SLOPEBCSF(CHF_FRA(a_dW),
                         CHF_CONST_FRA(a_W),
                         CHF_CONST_REAL(m_dx),
                         CHF_CONST_INT(a_dir),
                         CHF_BOX(loBox),
                         CHF_CONST_INT(hasLo),
                         CHF_BOX(hiBox),
                         CHF_CONST_INT(hasHi));
        }
    }
}

// Set up initial conditions
void GaussianIBC::initialize(LevelData<FArrayBox>& a_U)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  DataIterator dit = a_U.boxLayout().dataIterator();
  
  // Iterator of all grids in this level
  for (dit.begin(); dit.ok(); ++dit)
  {
    // Storage for current grid
    FArrayBox& U = a_U[dit()];

    // Box of current grid
    Box uBox = U.box();
    uBox &= m_domain;
    
    // Set up initial condition in this grid
    FORT_GAUSSIANINITF(CHF_CONST_FRA(U),
                       CHF_CONST_REAL(m_dx),
                       CHF_BOX(uBox));
  }
}
