/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// NodeFArrayBox.cpp
// from dMartin/Chombo-IAMR/util/FluxBox.cpp by Dan Martin, Fri, Jan 14, 2000
// petermc, 1 Nov 2000
// petermc, 26 Mar 2002 added using std::cout and using std::endl

#include "NodeFArrayBox.H"
#include "SPMD.H"
#include "DebugDump.H"
using std::cout;
using std::cerr;
using std::endl;

// first do simple access functions

// ---------------------------------------------------------
const Box&
NodeFArrayBox::box() const
{
  return m_box;
}

// ---------------------------------------------------------
FArrayBox& 
NodeFArrayBox::getFab()
{
  return m_fab;
}

// ---------------------------------------------------------
const FArrayBox& 
NodeFArrayBox::getFab() const
{
  return m_fab;
}


// ---------------------------------------------------------
// constructors and destructors
// ---------------------------------------------------------
// ": mfab()" means initialize m_fab with default constructor.
NodeFArrayBox::NodeFArrayBox() : m_fab()
{
}

// ---------------------------------------------------------
NodeFArrayBox::NodeFArrayBox(const Box& a_bx, int a_nComp)  : m_fab()
{
  define(a_bx, a_nComp);
}

// ---------------------------------------------------------
NodeFArrayBox::~NodeFArrayBox()
{

}


// ---------------------------------------------------------
// define function
void
NodeFArrayBox::define(const Box& a_bx, int a_nComp)
{
  assert(a_nComp > 0);
  assert(!a_bx.isEmpty());
  m_box = a_bx;
  Box nodeBox(surroundingNodes(a_bx));
  m_fab.define(nodeBox, a_nComp);
  
}
      

// ---------------------------------------------------------
void 
NodeFArrayBox::copy(const NodeFArrayBox& a_src)
{
  m_fab.copy(a_src.m_fab);
}

// ---------------------------------------------------------
void 
NodeFArrayBox::copy(const Box& a_regionFrom,
                    const Interval& a_Cdest,
                    const Box& a_regionTo,
                    const NodeFArrayBox& a_src,
                    const Interval& a_Csrc)
{
  assert(a_regionFrom.sameSize(a_regionTo));
  IntVect translator(a_regionTo.smallEnd() - a_regionFrom.smallEnd());
  const FArrayBox& srcFab = a_src.m_fab;
  // srcFab.box() should contain surroundingNodes(a_R)
  // assert(srcFab.box().contains(surroundingNodes(a_R)));
  // Box copyBox = surroundingNodes(a_R);
  Box nodesFrom = surroundingNodes(a_regionFrom);
  assert(srcFab.box().contains(nodesFrom));
  Box nodesTo = surroundingNodes(a_regionTo);
  // assert(m_fab.box().contains(nodesTo));
  // added by petermc, 15 Nov 2002, based on FluxBox::copy:
  // safety check -- due to node centering,
  // nodesTo may not be contained in m_fab.box().
  Box thisBox = m_fab.box();
  nodesTo &= thisBox;
  // Also need this so that nodesTo and nodesFrom are same size.
  nodesFrom &= srcFab.box(); // previously &= thisBox 
  Box nodesToTranslated(nodesTo);
  nodesToTranslated.shift(-translator);
  nodesFrom &= nodesToTranslated;
  m_fab.copy(nodesFrom, a_Cdest, nodesTo, srcFab, a_Csrc);
}


// ---------------------------------------------------------
int 
NodeFArrayBox::size(const Box& a_R, const Interval& a_comps) const
{
  Box nodeBox(surroundingNodes(a_R));
  return m_fab.size(nodeBox, a_comps);
}


// ---------------------------------------------------------
void
NodeFArrayBox::linearOut(void* a_buf, const Box& a_R, const Interval& a_comps) const
{
  Box nodeBox(surroundingNodes(a_R));
  // added by petermc, 5 Dec 2002, to avoid problems with copyTo().
  Box thisBox = m_fab.box();
  nodeBox &= thisBox;
  m_fab.linearOut(a_buf, nodeBox, a_comps);
}


// ---------------------------------------------------------
void 
NodeFArrayBox::linearIn(void* a_buf, const Box& a_R, const Interval& a_comps)
{
  Box nodeBox(surroundingNodes(a_R));
  // added by petermc, 5 Dec 2002, to avoid problems with copyTo().
  Box thisBox = m_fab.box();
  nodeBox &= thisBox;
  m_fab.linearIn(a_buf, nodeBox, a_comps);
}
