/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// NodeSetOperations.cpp
// petermc, 12 June 2003

#include "NodeSetOperations.H"
#include "LayoutIterator.H"
#include "BoxIterator.H"
using std::cout;
using std::cerr;
using std::endl;

// ---------------------------------------------------------
void
interiorNodes(IntVectSet& a_ivs,
              const Box& a_base_domain,
              const DisjointBoxLayout& a_boxLayout,
              const Box& a_box)
{
  ProblemDomain probdomain(a_base_domain);
  interiorNodes(a_ivs, probdomain, a_boxLayout, a_box);
}


// ---------------------------------------------------------
void
interiorNodes(IntVectSet& a_ivs,
              const ProblemDomain& a_base_domain,
              const DisjointBoxLayout& a_boxLayout,
              const Box& a_box)
{
  // define bgrow as the cells of a_box plus one ghost layer.
  // IntVectSet bgrow(grow(a_box, 1));
  Box bgrow(a_box);
  bgrow.grow(1);
  // intersect bgrow with the base domain.
  // (no effect in directions where base domain is periodic)
  bgrow &= a_base_domain;
  // shell is bgrow with all boxes at this level removed.
  IntVectSet shell(bgrow);

  Box periodicTestBox(a_base_domain.domainBox());
  if (a_base_domain.isPeriodic())
    for (int idir = 0; idir < SpaceDim; idir++)
      if (a_base_domain.isPeriodic(idir))
        periodicTestBox.grow(idir, -1);

  IntVect shiftMult = a_base_domain.domainBox().size();

  // (use LayoutIterator to get _all_ boxes, even those in other procs)
  LayoutIterator lit = a_boxLayout.layoutIterator();
  for (lit.begin(); lit.ok(); ++lit)
    {
      Box shiftedBox = a_boxLayout.get(lit());
      shell -= shiftedBox;

      // only do this IF we're periodic _and_ both boxes
      // adjoin the domain box boundary somewhere
      if (a_base_domain.isPeriodic() &&
          !periodicTestBox.contains(bgrow) &&
          !periodicTestBox.contains(shiftedBox))
        {
          ShiftIterator shiftIt = a_base_domain.shiftIterator();
          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
            {
              IntVect shiftVect(shiftMult*shiftIt());
              shiftedBox.shift(shiftVect);
              shell -= shiftedBox;
              shiftedBox.shift(-shiftVect);
            }
        }
    }
  // define interiorPlus as the cells removed from bgrow to yield shell.
  // const IntVectSet interiorPlus(bgrow - shell);
  IntVectSet interiorPlus(bgrow);
  interiorPlus -= shell;
  // initialize interior as interiorPlus.
  a_ivs.define(interiorPlus);
  Box offsetBox(IntVect::Zero, IntVect::Unit);
  // intersect a_ivs with interiorPlus with offsets
  BoxIterator offset(offsetBox);
  for (offset.begin(); offset.ok(); ++offset)
    {
      IntVectSet interiorPlus_off(interiorPlus);
      interiorPlus_off.shift(offset());
      a_ivs &= interiorPlus_off;
    }
}


// ---------------------------------------------------------
void
interiorBoundaryNodes(LayoutData< Vector<IntVectSet> >& a_IVSV,
                      const DisjointBoxLayout& a_boxes,
                      const Box& a_domain)
{
  ProblemDomain probdomain(a_domain);
  interiorBoundaryNodes(a_IVSV, a_boxes, probdomain);
}


// ---------------------------------------------------------
void
interiorBoundaryNodes(LayoutData< Vector<IntVectSet> >& a_IVSV,
                      const DisjointBoxLayout& a_boxes,
                      const ProblemDomain& a_domain)
{
  // boxesNodes contains the indices of the nodes of the boxes in a_boxes,
  // but they're represented as CELLs.
  DataIterator dit = a_boxes.dataIterator();
  // BoxLayout boxesNodes;
  // boxesNodes.deepCopy(a_boxes);
  // for (dit.begin(); dit.ok(); ++dit)
  // {
  // Box bx = a_boxes[dit()];
  // for (int idir = 0; idir < SpaceDim; idir++)
  // bx.growHi(idir, 1);
  // boxesNodes[dit()] = bx;
  // }
  // boxesNodes.close();

  // periodicTestBox is the domain box with one layer of cells removed
  // from each end in periodic directions.
  Box periodicTestBox(a_domain.domainBox());
  if (a_domain.isPeriodic())
    for (int idir = 0; idir < SpaceDim; idir++)
      if (a_domain.isPeriodic(idir))
        periodicTestBox.grow(idir, -1);

  // We use this to move things around in the periodic case.
  IntVect shiftMult = a_domain.domainBox().size();

  a_IVSV.define(a_boxes);
  for (dit.begin(); dit.ok(); ++dit)
    {
      Vector<IntVectSet>& IVSV = a_IVSV[dit()];

      const Box& bxCells = a_boxes.get(dit());

      // bxNodes is CELL-centered, contains indices of nodes of bxCells
      // const Box& bxNodes = boxesNodes.get(dit());

      // on each side, find interior boundary nodes of bxCells
      for (int idir = 0; idir < SpaceDim; idir++)
        {
          SideIterator sit;
          for (sit.begin(); sit.ok(); sit.next())
            {
              Side::LoHiSide mySide = sit();
              
              Box faceboxOriginal;
              switch (mySide)
                {
                case Side::Lo:
                  {
                    faceboxOriginal = adjCellLo(bxCells, idir, 1);
                    break;
                  }
                case Side::Hi:
                  {
                    faceboxOriginal = adjCellHi(bxCells, idir, 1);
                    break;
                  }
                default:
                  {
                    cerr << "interiorBoundaryNodes(): bogus side" << endl;
                    abort();
                  }
                }
              
              Box faceboxIn = faceboxOriginal & a_domain;
              // empty faceboxIn means physical boundary
              IntVectSet thisIVS(faceboxIn);
              Box facegrow(faceboxIn);
              if (! faceboxIn.isEmpty())
                {
                  // facegrow (CELL-centered) = faceboxIn expanded by:
                  // one layer in idir, side INTO the box;
                  // one layer in all non-idir directions;
                  // and then truncated to physical domain.
                  switch (mySide)
                    {
                    case Side::Lo:
                      {
                        facegrow.growHi(idir, 1);
                        break;
                      }
                    case Side::Hi:
                      {
                        facegrow.growLo(idir, 1);
                        break;
                      }
                    default:
                      {
                        cerr << "interiorBoundaryNodes(): bogus side" << endl;
                        abort();
                      }
                    }
                  
                  for (int idirOther = 0; idirOther < SpaceDim; idirOther++)
                    if (idirOther != idir) facegrow.grow(idirOther, 1);
                  
                  facegrow &= a_domain;

                  // box and facegrow for idir==0, mySide==Side::Hi
                  //
                  //              +---+
                  //      +-------+-+ |
                  //      |       | | |
                  //      |       | | |
                  //      |       | | |
                  //      +-------+-+ |
                  //              +---+
                  //

                  // shell is facegrow with all boxes on this level removed.
                  IntVectSet shell(facegrow);
                  // (use LayoutIterator to get _all_ boxes, even those in other procs)
                  LayoutIterator lit = a_boxes.layoutIterator();
                  for (lit.begin(); lit.ok(); ++lit)
                    {
                      Box shiftedBox = a_boxes.get(lit());
                      shell -= shiftedBox;
                      // added by petermc, 4 Feb 2003
                      // only do this IF we're periodic _and_ both boxes
                      // adjoin the domain box boundary somewhere
                      if (a_domain.isPeriodic() &&
                          !periodicTestBox.contains(facegrow) &&
                          !periodicTestBox.contains(shiftedBox))
                        {
                          ShiftIterator shiftIt = a_domain.shiftIterator();
                          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                            {
                              IntVect shiftVect(shiftMult*shiftIt());
                              shiftedBox.shift(shiftVect);
                              shell -= shiftedBox;
                              shiftedBox.shift(-shiftVect);
                            }
                        }
                    }

                  // define interiorPlus as the cells removed from facegrow
                  // to yield shell.
                  // const IntVectSet interiorPlus(facegrow - shell);
                  IntVectSet interiorPlus(facegrow);
                  interiorPlus -= shell;
                  // initialize interior as interiorPlus.
                  thisIVS = interiorPlus;

                  Box offsetBox(IntVect::Zero, IntVect::Unit);
                  // intersect thisIVS with interiorPlus with offsets
                  for (BoxIterator offset(offsetBox); offset.ok(); ++offset)
                    {
                      IntVectSet interiorPlus_off(interiorPlus);
                      interiorPlus_off.shift(offset());
                      thisIVS &= interiorPlus_off;
                    }
                }

              if (! thisIVS.isEmpty() )
                {
                  // IntVectSet IVS(bxNodes);
                  // petermc, 16 July 2003:  bxNodes is too big and forces
                  // a TreeIntVectSet.
                  IntVectSet IVS(facegrow);
                  IVS &= thisIVS;
                  // remove nodes that are already present in other components
                  for (int lcomp = 0; lcomp < IVSV.size(); lcomp++)
                    IVS -= IVSV[lcomp];
                  // add the new nodes as a new component of IVSV
                  if (! IVS.isEmpty() ) IVSV.push_back(IVS);
                }
            }
        }
    }
}


// ---------------------------------------------------------
void
interiorBoundaryNodes(LayoutData< Vector<IntVectSet> >& a_IVSV,
                      const DisjointBoxLayout& a_dest,
                      const DisjointBoxLayout& a_src,
                      const Box& a_domain)
{
  ProblemDomain probdomain(a_domain);
  interiorBoundaryNodes(a_IVSV, a_dest, a_src, probdomain);
}


// ---------------------------------------------------------
void
interiorBoundaryNodes(LayoutData< Vector<IntVectSet> >& a_IVSV,
                      const DisjointBoxLayout& a_dest,
                      const DisjointBoxLayout& a_src,
                      const ProblemDomain& a_domain)
{
  // destNodes contains the indices of the nodes of the boxes in a_dest,
  // but they're represented as cells.
  DataIterator dit = a_dest.dataIterator();
  BoxLayout destNodes;
  destNodes.deepCopy(a_dest);
  for (dit.begin(); dit.ok(); ++dit)
    {
      Box bx = a_dest[dit()];
      for (int idir = 0; idir < SpaceDim; idir++)
        bx.growHi(idir, 1);
      destNodes[dit()] = bx;
    }
  destNodes.close();

  Box periodicTestBox(a_domain.domainBox());
  if (a_domain.isPeriodic())
    for (int idirPeriodic = 0; idirPeriodic < SpaceDim; idirPeriodic++)
      if (a_domain.isPeriodic(idirPeriodic))
        periodicTestBox.grow(idirPeriodic, -1);

  IntVect shiftMult = a_domain.domainBox().size();

  // destSrcIntersec[dit()] is true iff the current source box
  // shares nodes with a_dest[dit()].
  LayoutData<bool> destSrcIntersect(a_dest);

  // cout << "interiorBoundaryNodes:  from " << a_src.size()
  //      << " boxes to " << a_dest.size() << " boxes"
  //      << " on " << a_domain.longside() << endl;
  // Define a_IVSV to have a Vector<IntVectSet> for each box of a_dest.
  a_IVSV.define(a_dest);
  for (LayoutIterator lit = a_src.layoutIterator(); lit.ok(); ++lit)
    {
      const Box& srcbox = a_src.get(lit());

      // srcNodesBox is cell-centered, contains indices of nodes of srcbox
      Box srcNodesBox = srcbox;
      for (int idir = 0; idir < SpaceDim; idir++)
        srcNodesBox.growHi(idir, 1);

      for (dit.begin(); dit.ok(); ++dit)
        {
          Box destNodesBox(destNodes.get(dit()));
          destNodesBox &= srcNodesBox;
          destSrcIntersect[dit()] = (! destNodesBox.isEmpty() );
        }
      
      // on each side, find interior boundary nodes of srcbox
      for (int idir = 0; idir < SpaceDim; idir++)
        {
          SideIterator sit;
          for (sit.begin(); sit.ok(); sit.next())
            {
              Side::LoHiSide mySide = sit();
              
              Box faceboxOriginal;
              switch (mySide)
                {
                case Side::Lo:
                  {
                    faceboxOriginal = adjCellLo(srcbox, idir, 1);
                    break;
                  }
                case Side::Hi:
                  {
                    faceboxOriginal = adjCellHi(srcbox, idir, 1);
                    break;
                  }
                default:
                  {
                    cerr << "interiorBoundaryNodes(): bogus side" << endl;
                    abort();
                  }
                }
              
              Box faceboxIn = faceboxOriginal & a_domain;
              // empty faceboxIn means physical boundary
              IntVectSet srcIVS(faceboxIn);
              Box facegrow(faceboxIn);
              if (! faceboxIn.isEmpty())
                {
                  // facegrow = faceboxIn expanded by:
                  // one layer in idir, side INTO the box;
                  // one layer in all non-idir directions;
                  // and then truncated to physical domain.
                  switch (mySide)
                    {
                    case Side::Lo:
                      {
                        facegrow.growHi(idir, 1);
                        break;
                      }
                    case Side::Hi:
                      {
                        facegrow.growLo(idir, 1);
                        break;
                      }
                    default:
                      {
                        cerr << "interiorBoundaryNodes(): bogus side" << endl;
                        abort();
                      }
                    }
                  
                  for (int idirOther = 0; idirOther < SpaceDim; idirOther++)
                    if (idirOther != idir) facegrow.grow(idirOther, 1);
                  
                  facegrow &= a_domain;

                  // box and facegrow for idir==0, mySide==Side::Hi
                  //
                  //              +---+
                  //      +-------+-+ |
                  //      |       | | |
                  //      |       | | |
                  //      |       | | |
                  //      +-------+-+ |
                  //              +---+
                  //

                  // shell is facegrow with all boxes on this level removed.
                  IntVectSet shell(facegrow);
                  // (use LayoutIterator to get _all_ boxes, even those in other procs)
                  LayoutIterator lit2 = a_src.layoutIterator();
                  for (lit2.begin(); lit2.ok(); ++lit2)
                    {
                      Box shiftedBox = a_src.get(lit2());
                      shell -= shiftedBox;
                      // added by petermc, 4 Feb 2003
                      // only do this IF we're periodic _and_ both boxes
                      // adjoin the domain box boundary somewhere
                      if (a_domain.isPeriodic() &&
                          !periodicTestBox.contains(faceboxIn) &&
                          !periodicTestBox.contains(shiftedBox))
                        {
                          ShiftIterator shiftIt = a_domain.shiftIterator();
                          for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                            {
                              IntVect shiftVect(shiftMult*shiftIt());
                              shiftedBox.shift(shiftVect);
                              shell -= shiftedBox;
                              shiftedBox.shift(-shiftVect);
                            }
                        }

                    }

                  // define interiorPlus as the cells removed from facegrow
                  // to yield shell.
                  // const IntVectSet interiorPlus(facegrow - shell);
                  IntVectSet interiorPlus(facegrow);
                  interiorPlus -= shell;
                  // initialize interior as interiorPlus.
                  srcIVS = interiorPlus;

                  Box offsetBox(IntVect::Zero, IntVect::Unit);
                  // intersect srcIVS with interiorPlus with offsets
                  for (BoxIterator offset(offsetBox); offset.ok(); ++offset)
                    {
                      IntVectSet interiorPlus_off(interiorPlus);
                      interiorPlus_off.shift(offset());
                      srcIVS &= interiorPlus_off;
                    }
                }

              if (! srcIVS.isEmpty() )
                {
                  for (dit.begin(); dit.ok(); ++dit)
                    {
                      if (destSrcIntersect[dit()])
                        // get nodes of destination box
                        {
                          Box destNodesBox(destNodes.get(dit()));
                          destNodesBox &= srcNodesBox;
                          if (! destNodesBox.isEmpty() )
                            {
                              // petermc, 16 Jul 2003:  destNodesBox was
                              // too big and forced TreeIntVectSet.
                              // Replace these two lines with reverse order.
                              // IntVectSet IVS(destNodesBox);
                              // IVS &= srcIVS;
                              IntVectSet IVS(srcIVS);
                              IVS &= destNodesBox;
                              // IVS: nodes of destination box that are also
                              // interior boundary nodes of source box
                              Vector<IntVectSet>& IVSV = a_IVSV[dit()];
                              // remove nodes that are already present in other components
                              for (int lcomp = 0; lcomp < IVSV.size(); lcomp++)
                                IVS -= IVSV[lcomp];
                              // add the new nodes as a new component
                              if (! IVS.isEmpty() ) IVSV.push_back(IVS);
                            }
                        }
                    }
                }
            }
        }
    }
}



// ---------------------------------------------------------
void
exteriorBoundaryNodes(LayoutData< Vector<IntVectSet> >& a_exterior,
                      const LayoutData< Vector<IntVectSet> >& a_interior,
                      const DisjointBoxLayout& a_boxes)
{
  a_exterior.define(a_boxes);
  for (DataIterator dit = a_boxes.dataIterator(); dit.ok(); ++dit)
    {
      Box box = a_boxes.get(dit());
      Vector<IntVectSet> ibn = a_interior[dit()];
      Vector<IntVectSet>& ebn = a_exterior[dit()];
      // on each side, find exterior boundary nodes
      for (int idir = 0; idir < SpaceDim; idir++)
        {
          SideIterator sit;
          for (sit.begin(); sit.ok(); sit.next())
            {
              Side::LoHiSide mySide = sit();
              
              // faceNodes will contain the cells with indices
              // that are the same as the nodes on this face.
              Box faceNodes;
              switch (mySide)
                {
                case Side::Lo:
                  {
                    faceNodes = adjCellLo(box, idir, 1);
                    // On low side, shift up by unit vector in direction idir,
                    // in order to get indices of the correct nodes.
                    faceNodes.shift(idir, 1);
                    break;
                  }
                case Side::Hi:
                  {
                    faceNodes = adjCellHi(box, idir, 1);
                    // On high side, the node indices are the same as
                    // the cell indices.
                    break;
                  }
                default:
                  {
                    cerr << "exteriorBoundaryNodes(): bogus side" << endl;
                    abort();
                  }
                }
              // Expand by one in all directions other than idir.
              for (int idirOther = 0; idirOther < SpaceDim; idirOther++)
                if (idirOther != idir) faceNodes.growHi(idirOther, 1);

              // IVS will be the face nodes without interior boundary nodes.
              // What's left will be the exterior boundary nodes.
              IntVectSet IVS(faceNodes);
              // remove the interior boundary nodes from IVS.
              for (int lcomp = 0; lcomp < ibn.size(); lcomp++)
                IVS -= ibn[lcomp];
              // add remaining nodes as a new component of
              // exterior boundary nodes.
              if (! IVS.isEmpty() ) ebn.push_back(IVS);
            }
        }
    }
}


// ---------------------------------------------------------
void 
zeroBoundaryNodes(BoxLayoutData<NodeFArrayBox>& a_dest,
                  const LayoutData< Vector<IntVectSet> >& a_IVSV)
{
  int numComps = a_dest.nComp();
  for (DataIterator dit = a_dest.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& destFab = a_dest[dit()].getFab();
      Vector<IntVectSet> IVS = a_IVSV[dit()];
      for (int lcomp = 0; lcomp < IVS.size(); lcomp++)
        for (IVSIterator it(IVS[lcomp]); it.ok(); ++it)
          { 
            IntVect iv = it();
            for (int ivar = 0; ivar < numComps; ivar++)
              destFab(iv, ivar) = 0.;
          }
    }
}


// ---------------------------------------------------------
void 
copyInteriorNodes(LevelData<NodeFArrayBox>& a_dest,
                  const LevelData<NodeFArrayBox>& a_src,
                  const LayoutData< Vector<IntVectSet> >& a_IVSV)
{
  const DisjointBoxLayout& destBoxes = a_dest.getBoxes();
  const DisjointBoxLayout& srcBoxes = a_src.getBoxes();
  assert(destBoxes.numBoxes(procID()) == a_IVSV.boxLayout().numBoxes(procID()));

  // (1) define psi on destBoxes
  // (2) copy a_src to psi;
  // (3) for each box in destBoxes
  //         for each interior node of srcBoxes in it,
  //             copy psi to a_dest there.

  // (1) define psi on destBoxes
  int numComps = a_src.nComp();
  BoxLayoutData<NodeFArrayBox> psi(destBoxes, numComps);
 
  // (2) copy a_src to psi;
  // petermc, 13 May 2003:  change Copier ghosts from Unit to Zero.
  Copier myCopier(srcBoxes, destBoxes, IntVect::Zero);
  a_src.copyTo(a_src.interval(), psi, psi.interval(), myCopier);

  // (3) for each box in destBoxes
  for (DataIterator dit = destBoxes.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& psiFab = psi[dit()].getFab();
      FArrayBox& destFab = a_dest[dit()].getFab();

      // copy psi to a_dest there.

      // destNodes:  indices of _all_ nodes in destBoxes[dit()]
      Box destNodes = destBoxes.get(dit());
      for (int dir = 0; dir < SpaceDim; dir++)
        destNodes.growHi(dir, 1);

      // (1) interior of each source box

      for (LayoutIterator lit = srcBoxes.layoutIterator(); lit.ok(); ++lit)
        {
          const Box thisBox = srcBoxes.get(lit());
          Box inner(thisBox);
          for (int dir = 0; dir < SpaceDim; dir++)
            inner.growLo(dir, -1);

          inner &= destNodes;

          // destFab.copy(psiFab, inner);  // incompatible box types
          for (BoxIterator bit(inner); bit.ok(); ++bit)
            { 
              IntVect iv = bit();
              for (int ivar = 0; ivar < numComps; ivar++)
                destFab(iv, ivar) = psiFab(iv, ivar);
            }
        }

      // (2) interior boundary points of source boxes

      Vector<IntVectSet> IVS = a_IVSV[dit()];
      for (int lcomp = 0; lcomp < IVS.size(); lcomp++)
        for (IVSIterator it(IVS[lcomp]); it.ok(); ++it)
          { 
            IntVect iv = it();
            for (int ivar = 0; ivar < numComps; ivar++)
              destFab(iv, ivar) = psiFab(iv, ivar);
          }
    }
}


// ---------------------------------------------------------
void
fullIntVectSets(LayoutData< BitSet >& a_IVSVfull,
                const LayoutData< Vector<IntVectSet> >& a_IVSV)
{
  const BoxLayout& grids = a_IVSV.boxLayout();
  a_IVSVfull.define(grids);
  for (DataIterator dit = grids.dataIterator(); dit.ok(); ++dit)
    {
      const Vector<IntVectSet>& IVSvec = a_IVSV[dit()];
      int ncomps = IVSvec.size();
      BitSet& fullvec = a_IVSVfull[dit()];
      fullvec = BitSet(ncomps, false);
      for (int lcomp = 0; lcomp < ncomps; lcomp++)
        {
          const IntVectSet& IVS = IVSvec[lcomp];
          assert(!IVS.isEmpty()); // shouldn't be empty, should it?
          Box container(IVS.minBox());
          if (container.numPts() == IVS.numPts()) fullvec.setTrue(lcomp);
        }
    }
}
