/* _______              __ 
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::ifstream;
using std::ios;

#include "FABView.H"

#include "ParmParse.H"
#include "CH_HDF5.H"
#include "parstream.H"
#include "FABView.H"

#include "AMR.H"
#include "AMRLevel.H"
#include "AMRLevelPolytropicGasFactory.H"
#include "AMRLevelPolytropicGas.H"

#include "PolytropicPhysics.H"

#include "RampIBC.H"
#include "ExplosionIBC.H"

//#define HALEM_PROC_SPEED
#ifdef HALEM_PROC_SPEED
#include <cstdio>
#include <sys/sysinfo.h>
#include <machine/hal_sysinfo.h>
#endif

#include "memusage.H"
#include "Timer.H"
Timer Everything    ("gov Everything", 0);
Timer TimeReadInput ("gov Read Input",   Everything);
Timer TimeSetupAMR  ("gov Setup AMR",    Everything);
Timer TimeRun       ("gov Run",          Everything);
Timer TimeConclude  ("gov Conclude",     Everything);

// amrGodunov is a function (as opposed to inline in main()) to get
// around MPI scoping problems
void amrGodunov();

// setupFixedGrids allows fixed grids to be read in and used in this AMR
// computation example
void setupFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                     const ProblemDomain&  a_domain,
                     int                   a_maxLevel,
                     int                   a_maxGridSize,
                     int                   a_blockFactor,
                     int                   a_verbosity,
                     std::string           a_gridFile);

// One more function for MPI
void dumpmemoryatexit();

int main(int a_argc, char* a_argv[])
{
#ifdef MPI
  // Start MPI
  MPI_Init(&a_argc,&a_argv);
#ifdef CH_AIX  
  H5dont_atexit();
#endif
  setChomboMPIErrorHandler();
#endif

  int rank, number_procs;
#ifdef MPI
  MPI_Comm_rank(Chombo_MPI::comm, &rank);
  MPI_Comm_size(Chombo_MPI::comm, &number_procs);
#else
  rank=0;
  number_procs=1;
#endif

  if (rank==0) {
    pout() << " number_procs= " << number_procs << endl;
  }

  Timer::TimerInit(rank);

  Everything.start();

  // Check for an input file
  char* inFile = NULL;

  if (a_argc > 1)
  {
    inFile = a_argv[1];
  }
  else
  {
    pout() << "Usage:  amrGodunov...ex <inputfile>" << endl;
    pout() << "No input file specified" << endl;
    return -1;
  }

  // Parse the command line and the input file (if any)
  ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

  // Run amrGodunov, i.e., do the computation
  amrGodunov();

  Everything.stop();

#ifdef TIMER
  Real end_memory = get_memory_usage();
  
  pout() << "Everything done.  mem= " << end_memory 
         << "MB wall-clock time: "    << setiosflags(ios::fixed)
         << Everything.wc_time()      << " seconds" << endl;
#endif

#if defined(TIMER) && defined(MPI)
  Real avg_memory, min_memory, max_memory;
  gather_memory_from_procs(end_memory, avg_memory, min_memory, max_memory);
#endif

  //Timer::TimerSummary();

#ifdef MPI
  // Exit MPI 
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}

void amrGodunov()
{
  
  TimeReadInput.start();
  // Read inputs that are prefixed with "godonouv."
  ParmParse ppgodunov("godunov");

  // This determines the amount of diagnositic output generated
  int verbosity = 0;
  ppgodunov.get("verbosity",verbosity);
  assert(verbosity >= 0);

  // For all gas dynamics
  Real gamma = 1.4;
  ppgodunov.get("gamma",gamma);

  // Stop after this number of steps
  int nstop = 0;
  ppgodunov.get("max_step",nstop);

  // Stop when the simulation time get here
  Real stopTime = 0.0;
  ppgodunov.get("max_time",stopTime);

  // Set the physical size of the longest dimension of the domain
  Real domainLength = 1.0;
  ppgodunov.get("domain_length",domainLength);

  // Set the resolution of the coarsest level 
  vector<int> numCells(SpaceDim);
  for (int i = 0; i < SpaceDim; ++i) numCells[i]=0;
  ppgodunov.getarr("num_cells",numCells,0,SpaceDim);

  assert(D_TERM(   (numCells[0] > 0),
                && (numCells[1] > 0),
                && (numCells[2] > 0)));
  assert(D_TERM(   (numCells[0] % 2 == 0),
                && (numCells[1] % 2 == 0),
                && (numCells[2] % 2 == 0)));

  // Determine which spatial directions are periodic
  vector<int> isPeriodica(SpaceDim,0);
  bool isPeriodic[SpaceDim];
  
  ppgodunov.getarr("is_periodic",isPeriodica,0,SpaceDim);
  // convert periodic from int->bool
  for (int dim=0; dim<SpaceDim; dim++) 
    {
      isPeriodic[dim] = (isPeriodica[dim] == 1);
      if (isPeriodic[dim] && verbosity >= 2 && procID() == 0)
        pout() << "Using Periodic BCs in direction: " << dim << endl;
    }

  // Maximum AMR level limit
  int maxLevel = 0;
  ppgodunov.get("max_level",maxLevel);
  int numReadLevels = Max(maxLevel,1);

  // Refinement ratios between levels
  std::vector<int> refRatios;
  // Note: this requires a refRatio to be defined for the finest level
  // (even though it will never be used)
  ppgodunov.getarr("ref_ratio",refRatios,0,numReadLevels+1);

  // Number of coarse time steps from one regridding to the next
  std::vector<int> regridIntervals;
  ppgodunov.getarr("regrid_interval",regridIntervals,0,numReadLevels);

  // How far to extend refinement from cells newly tagged for refinement
  int tagBufferSize = 3;
  ppgodunov.get("tag_buffer_size",tagBufferSize);

  // Threshold that triggers refinement
  Real refineThresh = 0.3;
  ppgodunov.get ("refine_thresh",refineThresh);

  // Minimum dimension of a grid
  int blockFactor = 1;
  ppgodunov.get("block_factor",blockFactor);

  // Maximum dimension of a grid
  int maxGridSize = 32;
  ppgodunov.get("max_grid_size",maxGridSize);

  Real fillRatio = 0.75;
  ppgodunov.get("fill_ratio",fillRatio);

  int inFlattening = 1;
  bool useFlattening;
  ppgodunov.get("use_flattening",inFlattening);
  useFlattening = (inFlattening == 1);

  int inCharLimiting = 0;
  bool useCharLimiting;
  ppgodunov.get("use_char_limiting",inCharLimiting);
  useCharLimiting = (inCharLimiting == 1);

  int inLimiting = 1;
  bool useLimiting;
  ppgodunov.get("use_limiting",inLimiting);
  useLimiting = (inLimiting == 1);

  // Apply artificial viscosity
  int inUseArtificialViscosity = 1;
  bool useArtificialViscosity;
  ppgodunov.get("use_artificial_viscosity",inUseArtificialViscosity);
  useArtificialViscosity = (inUseArtificialViscosity == 1);

  // Artificial viscosity coefficient/multiplier
  Real artificialViscosity = 0.1;
  ppgodunov.get("artificial_viscosity",artificialViscosity);

  // Set up checkpointing
  int checkpointInterval = 0;
  ppgodunov.get("checkpoint_interval",checkpointInterval);

  // Set up plot file writing
  int plotInterval = 0;
  ppgodunov.get("plot_interval",plotInterval);

  // CFL multiplier
  Real cfl = 0.8;
  ppgodunov.get("cfl",cfl);

  // Initial CFL multiplier
  Real initialCFL = 0.1;
  ppgodunov.get("initial_cfl",initialCFL);

  // Determine if a fixed or variable time step will be used
  Real fixedDt = -1;
  ppgodunov.get("fixed_dt",fixedDt);

  // Limit the time step growth
  Real maxDtGrowth = 1.1;
  ppgodunov.get("max_dt_growth",maxDtGrowth);

  // Let the time step grow by this factor above the "maximum" before
  // reducing it
  Real dtToleranceFactor = 1.1;
  ppgodunov.get("dt_tolerance_factor",dtToleranceFactor);

  TimeReadInput.stop();

#ifdef TIMER
   pout() << "Input Read. Beginning AMR Setup.  mem= " << get_memory_usage() 
          << "MB wall-clock time: " 
          << setiosflags(ios::fixed)
          << TimeReadInput.wc_time() 
          << " seconds" << endl;
#endif

  // Create initial and boundary condition (IBC) object and initialize
  TimeSetupAMR.start();
  PhysIBC* ibc;
  int whichproblem;
  ppgodunov.get("whichproblem", whichproblem);
  if(whichproblem == 0)
    {
      pout() << "Colella/Woodward mach snail problem " << endl;

      Real alpha, xcorner, ms;
      ppgodunov.get("angle_deg",alpha);
      ppgodunov.get("xcorner",xcorner);
      ppgodunov.get("shock_mach",ms);
      RampIBC* rampibc = new RampIBC(gamma,
                                     alpha,
                                     ms,
                                     xcorner,
                                     artificialViscosity);
      ibc = rampibc;
    }
  else if(whichproblem == 1)
  {
    pout() << "Explosion in a box problem " << endl;

    Real ms, explnsize;
    ppgodunov.get("shock_mach",ms);
    ppgodunov.get("explosion_size",explnsize);
    RealVect center;
    Vector<Real> vectCent(SpaceDim, 0.0);
    ppgodunov.getarr("explosion_center",vectCent,0,SpaceDim);

    for(int idir = 0; idir < SpaceDim; idir++) 
      center[idir] = vectCent[idir];

    ExplosionIBC* explosionibc = new ExplosionIBC(gamma,
                                                  ms,
                                                  center,
                                                  explnsize,
                                                  artificialViscosity);
    ibc = explosionibc;
  }

  ProblemDomain probDomain (IntVect::Zero,
                            IntVect(D_DECL(numCells[0]-1,
                                           numCells[1]-1,
                                           numCells[2]-1)),
                            isPeriodic);


  // Set up the patch integrator for polytropic gas dynamics
  PolytropicPhysics polytropicPhysics;
  polytropicPhysics.setPhysIBC(ibc);

  // Set up the AMRLevel... factory
  AMRLevelPolytropicGasFactory amrGodFact;
  GodunovPhysics* godunovPhysics = static_cast<GodunovPhysics*> (&polytropicPhysics);
  amrGodFact.CFL(cfl);
  amrGodFact.domainLength(domainLength);
  amrGodFact.refinementThreshold(refineThresh);
  amrGodFact.tagBufferSize(tagBufferSize);
  amrGodFact.godunovPhysics(godunovPhysics);
  amrGodFact.verbosity(verbosity);
  amrGodFact.initialDtMultiplier(initialCFL);
  amrGodFact.setSlopeParameters(useFlattening,
                                useLimiting,
                                useCharLimiting);
  amrGodFact.setArtificialViscosity(useArtificialViscosity,
                                    artificialViscosity);

  AMR amr;

  // Set up the AMR object
  amr.define(maxLevel,refRatios,probDomain,&amrGodFact);

  if (fixedDt > 0)
    {
      amr.fixedDt(fixedDt);
    }

  // Set grid generation parameters
  amr.maxGridSize(maxGridSize);
  amr.blockFactor(blockFactor);
  amr.fillRatio(fillRatio);

  // The hyperbolic codes use a grid buffer of 1
  amr.gridBufferSize(1);

  // Set output parameters
  amr.checkpointInterval(checkpointInterval);
  amr.plotInterval(plotInterval);
  amr.regridIntervals(regridIntervals);
  amr.maxDtGrow(maxDtGrowth);
  amr.dtToleranceFactor(dtToleranceFactor);

  // Set up output files
  if (ppgodunov.contains("plot_prefix"))
    {
      std::string prefix;
      ppgodunov.get("plot_prefix",prefix);
      amr.plotPrefix(prefix);
    }

  if (ppgodunov.contains("chk_prefix"))
    {
      std::string prefix;
      ppgodunov.get("chk_prefix",prefix);
      amr.checkpointPrefix(prefix);
    }

  amr.verbosity(verbosity);

  // Set up input files
  if (!ppgodunov.contains("restart_file"))
    {
      if (!ppgodunov.contains("fixed_hierarchy"))
        {
          // initialize from scratch for AMR run
          // initialize hierarchy of levels
          amr.setupForNewAMRRun();
        }
      else
        {
          std::string gridFile;
          ppgodunov.get("fixed_hierarchy",gridFile);

          // initialize from a list of grids in "gridFile"
          Vector<Vector<Box> > amrGrids(maxLevel+1);
          setupFixedGrids(amrGrids,
                          probDomain,
                          maxLevel,
                          maxGridSize,
                          blockFactor,
                          verbosity,
                          gridFile);
          amr.setupForFixedHierarchyRun(amrGrids,1);
        }
    }
  else
    {
      std::string restartFile;
      ppgodunov.get("restart_file",restartFile);

#ifdef HDF5
      HDF5Handle handle(restartFile,HDF5Handle::OPEN_RDONLY);
      // read from checkpoint file
      amr.setupForRestart(handle);
      handle.close();
#else
      MayDay::Error("amrGodunov restart only defined with hdf5");
#endif
    }

  TimeSetupAMR.stop();

#ifdef TIMER
   pout() << "AMR Setup done.  mem= " << get_memory_usage() 
          << "MB wall-clock time: " 
          << setiosflags(ios::fixed)
          << TimeSetupAMR.wc_time() 
          << " seconds" << endl;
#endif

  // Run the computation
  TimeRun.start();
  amr.run(stopTime,nstop);
  TimeRun.stop();

#ifdef TIMER
   pout() << "AMR Run completed.  mem= " << get_memory_usage() 
          << "MB wall-clock time: " 
          << setiosflags(ios::fixed)
          << TimeRun.wc_time() 
          << " seconds" << endl;
#endif

  // Output the last plot file and statistics
  TimeConclude.start();
  amr.conclude();
  TimeConclude.stop();

#ifdef TIMER
   pout() << "AMR Conclude completed.  mem= " << get_memory_usage() 
          << "MB wall-clock time: " 
          << setiosflags(ios::fixed)
          << TimeConclude.wc_time() 
          << " seconds" << endl;
#endif

}

// setupFixedGrids allows fixed grids to be read in and used in this AMR
// computation example
void setupFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                     const ProblemDomain&  a_domain,
                     int                   a_maxLevel,
                     int                   a_maxGridSize,
                     int                   a_blockFactor,
                     int                   a_verbosity,
                     std::string           a_gridFile)
{
  // Run this task on one processor
  if (procID() == uniqueProc(SerialTask::compute))
  {
    a_amrGrids.push_back(Vector<Box>(1,a_domain.domainBox()));

    // Read in predefined grids
    ifstream is(a_gridFile.c_str(), ios::in);

    if (is.fail())
    {
      MayDay::Error("Cannot open grids file");
    }
  
    // Format of file:
    //   number of levels, then for each level (starting with level 1):
    //   number of grids on level, list of boxes

    int inNumLevels;
    is >> inNumLevels;

    assert (inNumLevels <= a_maxLevel+1);

    if (a_verbosity >= 3)
    {
      pout() << "numLevels = " << inNumLevels << endl;
    }

    while (is.get() != '\n');

    a_amrGrids.resize(inNumLevels);

    // Check to see if coarsest level needs to be broken up
    domainSplit(a_domain,a_amrGrids[0],a_maxGridSize,a_blockFactor);

    if (a_verbosity >= 3)
    {
      pout() << "level 0: ";
      for (int n = 0; n < a_amrGrids[0].size(); n++)
      {
        pout() << a_amrGrids[0][0] << endl;
      }
    }

    // Now loop over levels, starting with level 1
    int ngrid;
    for (int lev = 1; lev < inNumLevels; lev++)
    {
      is >> ngrid;

      if (a_verbosity >= 3)
      {
        pout() << "level " << lev << " numGrids = " << ngrid << endl;
        pout() << "Grids: ";
      }

      while (is.get() != '\n');

      a_amrGrids[lev].resize(ngrid);

      for (int i = 0; i < ngrid; i++)
      {
        Box bx;
        is >> bx;

        while (is.get() != '\n');

        // Quick check on box size
        Box bxRef(bx);

        if (bxRef.longside() > a_maxGridSize)
        {
          pout() << "Grid " << bx << " too large" << endl;
          MayDay::Error();
        }

        if (a_verbosity >= 3) 
        {
          pout() << bx << endl;
        }

        a_amrGrids[lev][i] = bx;
      } // End loop over boxes on this level
    } // End loop over levels
  }

  // Broadcast results to all the processors
  broadcast(a_amrGrids,uniqueProc(SerialTask::compute));
}
