/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "PolytropicPhysics.H"
#include "PolytropicPhysicsF_F.H"
#include "LoHiCenter.H"
#include "LoHiSide.H"
#include "LGintegrator.H"

/////////
void 
PolytropicPhysics::getFlux(FArrayBox&       a_flux,
                           const FArrayBox& a_whalf,
                           const Box&       a_box,
                           const int&       a_idir)
{
  FORT_GETFLUX(CHF_FRA(a_flux),
               CHF_CONST_FRA(a_whalf),
               CHF_CONST_INT(a_idir),
               CHF_BOX(a_box));
}
/////////
PolytropicPhysics::PolytropicPhysics():GodunovPhysics()
{
}
/////////
PolytropicPhysics::~PolytropicPhysics()
{
}
/////////
GodunovPhysics* PolytropicPhysics::new_godunovPhysics() const
{
  assert(m_isBCSet);
  GodunovPhysics* retval = static_cast<GodunovPhysics*>(new PolytropicPhysics());

  retval->setPhysIBC(m_bc);
  return retval;
}

// Compute the maximum wave speed
Real 
PolytropicPhysics::getMaxWaveSpeed(const FArrayBox& a_U,
                                   const Box&       a_box)
{
  assert(isDefined());
  assert(a_U.contains(a_box));

  Real speed = 0.0;

  FORT_MAXWAVESPEED(CHF_REAL(speed),
                    CHF_CONST_FRA(a_U),
                    CHF_BOX(a_box));

  return speed;
}

// Number of conserved variables
int PolytropicPhysics::numConserved()
{
  return CNUM;
}

// Names of the conserved variables
Vector<string> PolytropicPhysics::stateNames()
{
  Vector<string> retval;

  retval.push_back("density");
  retval.push_back("x-momentum");
  if (SpaceDim >= 2)
    {
      retval.push_back("y-momentum");

      if (SpaceDim >= 3)
        {
          retval.push_back("z-momentum");
        }
    }
  retval.push_back("energy-density");
  return retval; 
}

// Number of flux variables
int PolytropicPhysics::numFluxes()
{
  // In some computations there may be more fluxes than conserved variables
  return CNUM;
}

// Number of primitive variables
int PolytropicPhysics::numPrimitives()
{
  // This doesn't equal the number of conserved variables because
  // auxiliary/redundant variable may be computed and stored
  return QNUM;
}

// Compute the primitive variables from the conserved variables
void PolytropicPhysics::consToPrim(FArrayBox&       a_W,
                                   const FArrayBox& a_U,
                                   const Box&       a_box)
{
  assert(isDefined());
  assert(a_U.box().contains(a_box));
  assert(a_W.box().contains(a_box));

  FORT_CONSTOPRIM(CHF_FRA(a_W),
                  CHF_CONST_FRA(a_U),
                  CHF_BOX(a_box));
}
/////////
// Compute expansion amplitudes of dW in right eigenvectors.
void PolytropicPhysics::charAnalysis(FArrayBox&       a_dW,
                                     const FArrayBox& a_W,
                                     const int&       a_dir,
                                     const Box&       a_box)
{
  assert(isDefined());

  FORT_CHAREXP(CHF_FRA(a_dW),
               CHF_CONST_FRA(a_W),
               CHF_CONST_INT(a_dir),
               CHF_BOX(a_box));
}
/////////
void PolytropicPhysics::charSynthesis(FArrayBox&       a_dW,
                                      const FArrayBox& a_W,
                                      const int&       a_dir,
                                      const Box&       a_box)
{
  assert(isDefined());

  FORT_CHARSYN(CHF_FRA(a_dW),
               CHF_CONST_FRA(a_W),
               CHF_CONST_INT(a_dir),
               CHF_BOX(a_box));
}

void 
PolytropicPhysics::charValues(FArrayBox&       a_lambda,
                              const FArrayBox& a_W,
                              const int&       a_dir,
                              const Box&       a_box)
{
  assert(isDefined());

  FORT_CHARVALS(CHF_FRA(a_lambda),
                CHF_CONST_FRA(a_W),
                CHF_CONST_INT(a_dir),
                CHF_BOX(a_box));
}

// Compute a Riemann problem and generate fluxes at the faces.

void PolytropicPhysics::riemann(FArrayBox&       a_WGdnv,
                                const FArrayBox& a_WLeft,
                                const FArrayBox& a_WRight,
                                const FArrayBox& a_W,
                                const Real&      a_time,
                                const int&       a_dir,
                                const Box&       a_box)
{
  assert(isDefined());

  assert(a_WGdnv.box().contains(a_box));

  // Get the numbers of relevant variables
  int numPrim = numPrimitives();

  assert(a_WGdnv.nComp() == numPrim);
  assert(a_WLeft.nComp() == numPrim);
  assert(a_WRight.nComp() == numPrim);

  // Cast away "const" inputs so their boxes can be shifted left or right
  // 1/2 cell and then back again (no net change is made!)
  FArrayBox& shiftWLeft  = (FArrayBox&)a_WLeft;
  FArrayBox& shiftWRight = (FArrayBox&)a_WRight;

  // Solution to the Riemann problem

  // Shift the left and right primitive variable boxes 1/2 cell so they are
  // face centered
  shiftWLeft .shiftHalf(a_dir, 1);
  shiftWRight.shiftHalf(a_dir,-1);
  
  assert(shiftWLeft.box().contains(a_box));
  assert(shiftWRight.box().contains(a_box));

  // Riemann solver computes Wgdnv all edges that are not on the physical
  // boundary.
  FORT_RIEMANNF(CHF_FRA(a_WGdnv),
                CHF_CONST_FRA(shiftWLeft),
                CHF_CONST_FRA(shiftWRight),
                CHF_CONST_INT(a_dir),
                CHF_BOX(a_box));
 
  // Shift the left and right primitive variable boxes back to their original
  // position.

  // Call boundary Riemann solver.
  m_bc->primBC(a_WGdnv,shiftWLeft,a_W,a_dir, Side::Hi, a_time);
  m_bc->primBC(a_WGdnv,shiftWRight,a_W,a_dir, Side::Lo, a_time);
  shiftWLeft .shiftHalf(a_dir,-1);
  shiftWRight.shiftHalf(a_dir, 1);
}

void PolytropicPhysics::postNormalPredictor(FArrayBox&       a_dWLow,
                                            FArrayBox&       a_dWHigh,
                                            const FArrayBox& a_W,
                                            const int&       a_dir,
                                            const Box&       a_box)

{
  // Not required for gas dynamics.
}
// /compute increment of primitive variables.
void PolytropicPhysics::quasilinearUpdate(FArrayBox&       a_dWdx,
                                          const FArrayBox& a_WHalf,
                                          const FArrayBox& a_W,
                                          const Real&      a_scale,
                                          const int&       a_dir,
                                          const Box&       a_box)
{
  assert(isDefined());
  assert(a_dWdx.box().contains(a_box));

  FORT_GETADWDX(CHF_FRA(a_dWdx),
                CHF_CONST_FRA(a_WHalf),
                CHF_CONST_FRA(a_W),
                CHF_CONST_REAL(a_scale),
                CHF_CONST_INT(a_dir),
                CHF_BOX(a_box));
}

// Interval within the primitive variables corresponding to the velocities
Interval 
PolytropicPhysics::velocityInterval()
{
#if CH_SPACEDIM==2
  Interval retval(QVELX,QVELY);
#elif CH_SPACEDIM==3
  Interval retval(QVELX,QVELZ);
#else 
  bogus_spacedim();
#endif

  return retval;
}
// Component index within the primitive variables of the pressure
int 
PolytropicPhysics::pressureIndex()
{
  return QPRES;
}
// Component index within the primitive variables of the bulk modulus
int 
PolytropicPhysics::bulkModulusIndex()
{
  return QPRES;
}
