/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DFMartin, Sun, May 5, 2002

#ifndef _BICGSTABSMOOTHER_H_
#define _BICGSTABSMOOTHER_H_

#include <iostream>
#include <cmath>
#include <cassert>
#include <cstdlib>

#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "ProblemDomain.H"

#include "BaseBottomSmoother.H"

/// Implements BiCGStab as a bottom smoother for LevelOp-derived classes
/**
    The BiCGStabSmoother class implements the BiCGStab algorithm for
    LevelOp-derived classes
*/
class BiCGStabSmoother: public BaseBottomSmoother
{
public:
  /// Null constructor
  BiCGStabSmoother();

  /// Destructor
  virtual ~BiCGStabSmoother();

  /// "Virtual constructor" workaround.
  /**
      This gets around the "no virtual constructor" rule
   */
  virtual BaseBottomSmoother* new_bottomSmoother() const;

  /// Set maximum number of BiCGStab iterations.
  void setMaxIter(int a_maxIter);

  /// Set solver tolerance
  /**
      Set solver tolerance 
   */
  void setSolverTol(Real a_solverTol);

  /// Set whether to enable restart capabillity
  /**
      If this is true, if BiCGStab reaches a hang point, the
      solver will recompute the residual and attempt to restart the
      computation
  */
  void setEnableRestart(bool a_enableRestart);

  /// Set verbosity level
  /**
      Set whether i/o is done when solving (default is false)
  */
  void setVerbose(bool a_verbose);

  /// Do smoothing
  /**
      Performs BiCGStab smoothing of L(a_phi) = a_rhs; L is the operator
      defined in a_levelopPtr, and the preconditioner is
      a_levelopPtr->smooth
  */
  virtual void doBottomSmooth(LevelData<FArrayBox>&       a_phi,
                              const LevelData<FArrayBox>& a_rhs,
                              LevelOp*                    a_levelopPtr);

  /// set metric for judging convergence (if not set, defaults to 1.0)
  virtual void setConvergenceMetric(Real a_metric, int a_comp);

protected:
  /**
     Maximum number of BiCGStab iterations
  */
  int m_maxIter;

  /** 
      Solver convergence metrics
  */
  Vector<Real> m_convergenceMetrics;

  /**
     Solver tolerance (we declare the problem solved if the residual is 
     smaller than solverTol*convergenceMetric
  */
  Real m_solverTol;

  /**
     Small parameter to check for division by zero
  */
  Real m_small;

  /**
     Small parameter to check for size of correction (to prevent hanging)
  */
  Real m_convergeSmall;

  /**
     If true, enable restarting capability
  */
  bool m_enableRestart;

  /**
     Default is false
  */
  bool m_verbose;
};

#endif
