/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LayoutIterator.H"
#include "DataIterator.H"

#include "CFIVS.H"

bool CFIVS::isDefined() const
{
  return m_isDefined;
} 

const IntVectSet& CFIVS::getFineIVS() const 
{
  assert(m_isDefined);

  return m_fiinterpIVS;
}

void CFIVS::setDefaultValues()
{
  m_fiinterpIVS.define();
  m_packed = false;

  m_isDefined = false;
}

CFIVS::CFIVS()
{
  setDefaultValues();
}

CFIVS::~CFIVS()
{
  setDefaultValues();
}

CFIVS::CFIVS(const Box&               a_domain, 
             const Box&               a_boxIn,
             const DisjointBoxLayout& a_fineBoxes, 
             int                      a_direction,
             Side::LoHiSide           a_hiorlo)
{
  setDefaultValues();
  ProblemDomain probdomain(a_domain);

  define(probdomain, a_boxIn, a_fineBoxes, a_direction, a_hiorlo);
           
}

CFIVS::CFIVS(const ProblemDomain&     a_domain, 
             const Box&               a_boxIn,
             const DisjointBoxLayout& a_fineBoxes, 
             int                      a_direction,
             Side::LoHiSide           a_hiorlo)
{
  setDefaultValues();

  define(a_domain, a_boxIn, a_fineBoxes, a_direction, a_hiorlo);
           
}

void CFIVS::define(const Box&               a_domain, 
                   const Box&               a_boxIn,
                   const DisjointBoxLayout& a_fineBoxes, 
                   int                      a_direction,
                   Side::LoHiSide           a_hiorlo)
{
  ProblemDomain probdomain(a_domain);

  define(probdomain, a_boxIn, a_fineBoxes, a_direction, a_hiorlo);
}

void CFIVS::define(const ProblemDomain&     a_domain, 
                   const Box&               a_boxIn,
                   const DisjointBoxLayout& a_fineBoxes, 
                   int                      a_direction,
                   Side::LoHiSide           a_hiorlo)
{
  m_isDefined = true;

  assert(a_direction >= 0);
  assert(a_direction < SpaceDim);
  assert(!a_domain.isEmpty());
  assert(a_domain.contains(a_boxIn));
  assert(a_fineBoxes.checkPeriodic(a_domain));
  assert((a_hiorlo == Side::Lo) || (a_hiorlo == Side::Hi));

  // shift direction
  int hilo = sign(a_hiorlo);
    
  // create fine stencil
  Box finebox = a_boxIn;
  Box edgebox;

  assert((hilo ==1) || (hilo == -1));

  if (hilo == -1)
    {
      edgebox = adjCellLo(finebox,a_direction,1);
    }
  else
    {
      edgebox = adjCellHi(finebox,a_direction,1);
    }

  edgebox &= a_domain;

  if (!edgebox.isEmpty())
    {
      m_fiinterpIVS.define(edgebox);

      LayoutIterator lit = a_fineBoxes.layoutIterator();
      Box periodicTestBox(a_domain.domainBox());

      if (a_domain.isPeriodic()) 
        {
          for (int idir = 0; idir < SpaceDim; idir++)
            {
              if (a_domain.isPeriodic(idir)) 
                periodicTestBox.grow(idir,-1);
            }
        }

      for (lit.reset(); lit.ok(); ++lit)
        {
          m_fiinterpIVS -= a_fineBoxes[lit()];

          // only do this IF we're periodic _and_ both boxes
          // adjoin the domain box boundary somewhere
          if (a_domain.isPeriodic() && !periodicTestBox.contains(edgebox)
              && !periodicTestBox.contains(a_fineBoxes[lit()]))
            {
              ShiftIterator shiftIt = a_domain.shiftIterator();
              IntVect shiftMult = a_domain.domainBox().size();
              Box shiftedBox = a_fineBoxes[lit()];

              for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)
                {
                  IntVect shiftVect(shiftMult*shiftIt());

                  shiftedBox.shift(shiftVect);
                  m_fiinterpIVS -= shiftedBox;
                  shiftedBox.shift(-shiftVect);
                }
            }
        }
    }

  m_fiinterpIVS.compact();

  if (m_fiinterpIVS.isEmpty())
    {
      m_packed = false;
    }
  else
  {
    Box min = m_fiinterpIVS.minBox();

    if (min.numPts() == m_fiinterpIVS.numPts())
    {
      m_packCount++;
      m_packed = true;
      m_packedBox = min;
    }
    else
    {
      m_sparseCount++;
      m_packed = false;
    }
  }
}

long long CFIVS::m_packCount = 0;

long long CFIVS::m_sparseCount = 0;
