/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _FORT_PROTO_H_
#define _FORT_PROTO_H_

#include "REAL.H"
#include "SPACE.H"

#ifdef DOXYGEN
/// Used to declare a fortran routine in a portable way 
/** @ingroup macros

 */
#define FORTRAN_NAME( NAME ,name ) 

/// Used in ChomboFortran to write dimension-indenpendent loops
/** @ingroup macros
CHF_MULTIDO is used to iterate over a box in a dimension independent
fashion by setting up nested Fortran do loops and CHF_ENDDO is used to
terminate those do loops correctly.  Specifically,
CHF_MULTIDO[box;i;j;k] will generate a do loop for  i
nested inside a do loop for  j and, in 3D, this will be nested inside
a do loop for k.  The i loop will go from first element of the
low corner of box to the first element of the high corner of box.
Similarly, the {\tt j} loop will use the second element and, in 3D, the
 k loop will use the third element.  CHF_ENDDO will end all the do 
loops set up by  CHF_MULTIDO.

Here is an example using these macros:

\code
      subroutine LOOP(CHF_FRA1[array],CHF_BOX[box])

      integer CHF_DDECL[i;j;k]
      integer productsum

      productsum = 0
      CHF_MULTIDO[box;i;j;k]
        productsum = productsum + i*j*k
        array(CHF_IX[i;j;k]) = productsum
      CHF_ENDDO

      return
      end
\endcode

see the section of the Chombo Design document about Chombo Fortran for
a more detailed use of macros in Fortran

*/
#define CHF_MULTIDO(box, i, j, k)

/** @ingroup macros */
#define CHF_ENDDO

#else

#ifdef CH_FORT_UPPERCASE
  #ifdef CH_FORT_UNDERSCORE
     #define FORTRAN_NAME( NAME ,name ) NAME ## _
  #else
     #define FORTRAN_NAME( NAME ,name ) NAME
  #endif
#else
  #ifdef CH_FORT_UNDERSCORE
     #define FORTRAN_NAME( NAME ,name ) name ## _
  #else
     #define FORTRAN_NAME( NAME ,name ) name
  #endif
#endif

class chfptr {
public:
  int i ;
  //XXX for debugging
  //XXXchfptr(int a):i(a){std::cout << "constructing chfptr " << &i << std::endl;}
  //XXX~chfptr() {std::cout << "destructing chfptr " << &i << std::endl;}
  chfptr(int a):i(a){}
  operator const int*() const {return &i;}
};

#endif

// ChomboFortran declaration macros for Application code

// primitive data types

/** @ingroup macros */
#define CHF_INT( i ) &i
/** @ingroup macros */
#define CHF_CONST_INT( i ) &i
/** @ingroup macros */
#define CHF_REAL( r ) &r
/** @ingroup macros */
#define CHF_CONST_REAL( r ) &r
/** @ingroup macros */
#define CHF_I1D(       ia,dimen ) ia,(const int*)chfptr(dimen-1)
/** @ingroup macros */
#define CHF_CONST_I1D( ia,dimen ) ia,(const int*)chfptr(dimen-1)
/** @ingroup macros */
#define CHF_R1D(       ra,dimen ) ra,(const int*)chfptr(dimen-1)
/** @ingroup macros */
#define CHF_CONST_R1D( ra,dimen ) ra,(const int*)chfptr(dimen-1)

// Chombo data types

/** @ingroup macros */
#define CHF_INTVECT( a ) a.dataPtr( ) 
/** @ingroup macros */
#define CHF_CONST_INTVECT( a ) a.dataPtr( ) 
/** @ingroup macros */
#define CHF_REALVECT( a ) a.dataPtr( ) 
/** @ingroup macros */
#define CHF_CONST_REALVECT( a ) a.dataPtr( ) 
/** @ingroup macros */
#define CHF_VI(       v ) &v[0],(const int*)chfptr(v.size()-1)
/** @ingroup macros */
#define CHF_CONST_VI( v ) &v[0],(const int*)chfptr(v.size()-1)
/** @ingroup macros */
#define CHF_VR(       v ) &v[0],(const int*)chfptr(v.size()-1)
/** @ingroup macros */
#define CHF_CONST_VR( v ) &v[0],(const int*)chfptr(v.size()-1)

/** @ingroup macros */
#define CHF_BOX( b ) D_DECL( &b.loVect()[0] \
                            ,&b.loVect()[1] \
                            ,&b.loVect()[2] ) \
                    ,D_DECL( &b.hiVect()[0] \
                            ,&b.hiVect()[1] \
                            ,&b.hiVect()[2] ) 

/** @ingroup macros */
#define CHF_FIA( a ) a.dataPtr( 0 ) \
                     ,D_DECL( &a.loVect()[0] \
                             ,&a.loVect()[1] \
                             ,&a.loVect()[2] ) \
                     ,D_DECL( &a.hiVect()[0] \
                             ,&a.hiVect()[1] \
                             ,&a.hiVect()[2] ) \
                     ,a.nCompPtr()

/** @ingroup macros */
#define CHF_FRA( a ) a.dataPtr( 0 ) \
                     ,D_DECL( &a.loVect()[0] \
                             ,&a.loVect()[1] \
                             ,&a.loVect()[2] ) \
                     ,D_DECL( &a.hiVect()[0] \
                             ,&a.hiVect()[1] \
                             ,&a.hiVect()[2] ) \
                     ,a.nCompPtr()

/** @ingroup macros */
#define CHF_CONST_FIA( a ) a.dataPtr( 0 ) \
                     ,D_DECL( &a.loVect()[0] \
                             ,&a.loVect()[1] \
                             ,&a.loVect()[2] ) \
                     ,D_DECL( &a.hiVect()[0] \
                             ,&a.hiVect()[1] \
                             ,&a.hiVect()[2] ) \
                     ,a.nCompPtr()

/** @ingroup macros */
#define CHF_CONST_FRA( a ) a.dataPtr( 0 ) \
                     ,D_DECL( &a.loVect()[0] \
                             ,&a.loVect()[1] \
                             ,&a.loVect()[2] ) \
                     ,D_DECL( &a.hiVect()[0] \
                             ,&a.hiVect()[1] \
                             ,&a.hiVect()[2] ) \
                     ,a.nCompPtr()

/** @ingroup macros */
#define CHF_FIA1( a,n ) a.dataPtr( n ) \
                     ,D_DECL( &a.loVect()[0] \
                             ,&a.loVect()[1] \
                             ,&a.loVect()[2] ) \
                     ,D_DECL( &a.hiVect()[0] \
                             ,&a.hiVect()[1] \
                             ,&a.hiVect()[2] )

/** @ingroup macros */
#define CHF_FRA1( a,n ) a.dataPtr( n ) \
                     ,D_DECL( &a.loVect()[0] \
                             ,&a.loVect()[1] \
                             ,&a.loVect()[2] ) \
                     ,D_DECL( &a.hiVect()[0] \
                             ,&a.hiVect()[1] \
                             ,&a.hiVect()[2] )

/** @ingroup macros */
#define CHF_CONST_FIA1( a,n ) a.dataPtr( n ) \
                     ,D_DECL( &a.loVect()[0] \
                             ,&a.loVect()[1] \
                             ,&a.loVect()[2] ) \
                     ,D_DECL( &a.hiVect()[0] \
                             ,&a.hiVect()[1] \
                             ,&a.hiVect()[2] )

/** @ingroup macros */
#define CHF_CONST_FRA1( a,n ) a.dataPtr( n ) \
                     ,D_DECL( &a.loVect()[0] \
                             ,&a.loVect()[1] \
                             ,&a.loVect()[2] ) \
                     ,D_DECL( &a.hiVect()[0] \
                             ,&a.hiVect()[1] \
                             ,&a.hiVect()[2] )


// Macros used in automatically-generated Chombo Fortran prototypes

// primitive data types

#define CHFp_INT( i ) int* const i
#define CHFp_CONST_INT( i ) const int* const i
#define CHFp_REAL( r ) Real* const r
#define CHFp_CONST_REAL( r ) const Real* const r
//[NOTE: the *1D macros has 2 args in C++, but just 1 in Fortran]
#define CHFp_I1D(       ia,dimen )        int* const ia, const int* const i ## ia ## hi0
#define CHFp_CONST_I1D( ia,dimen ) const  int* const ia, const int* const i ## ia ## hi0
#define CHFp_R1D(       ra,dimen )       Real* const ra, const int* const i ## ra ## hi0
#define CHFp_CONST_R1D( ra,dimen ) const Real* const ra, const int* const i ## ra ## hi0

// Chombo data types

#define CHFp_INTVECT( r ) int* const r
#define CHFp_CONST_INTVECT( r ) const int* const r
#define CHFp_REALVECT( r )  Real* const r
#define CHFp_CONST_REALVECT( r ) const Real* const r
#define CHFp_VI(       v )        int * const v, const int* const i ## v ## hi0
#define CHFp_CONST_VI( v ) const  int * const v, const int* const i ## v ## hi0
#define CHFp_VR(       v )       Real * const v, const int* const i ## v ## hi0
#define CHFp_CONST_VR( v ) const Real * const v, const int* const i ## v ## hi0

#define CHFp_BOX( b ) D_DECL( const int* const i ## b ## lo0 \
                             ,const int* const i ## b ## lo1 \
                             ,const int* const i ## b ## lo2 ) \
                     ,D_DECL( const int* const i ## b ## hi0 \
                             ,const int* const i ## b ## hi1 \
                             ,const int* const i ## b ## hi2 ) 

#define CHFp_FIA( a ) int *const a \
                     ,D_DECL( const int* const i ## a ## lo0 \
                             ,const int* const i ## a ## lo1 \
                             ,const int* const i ## a ## lo2 ) \
                     ,D_DECL( const int* const i ## a ## hi0 \
                             ,const int* const i ## a ## hi1 \
                             ,const int* const i ## a ## hi2 ) \
                             ,const int* const n ## a ## comp
#define CHFp_FRA( a ) Real *const a \
                     ,D_DECL( const int* const i ## a ## lo0 \
                             ,const int* const i ## a ## lo1 \
                             ,const int* const i ## a ## lo2 ) \
                     ,D_DECL( const int* const i ## a ## hi0 \
                             ,const int* const i ## a ## hi1 \
                             ,const int* const i ## a ## hi2 ) \
                             ,const int* const n ## a ## comp
#define CHFp_CONST_FIA( a ) const int *const a \
                     ,D_DECL( const int* const i ## a ## lo0 \
                             ,const int* const i ## a ## lo1 \
                             ,const int* const i ## a ## lo2 ) \
                     ,D_DECL( const int* const i ## a ## hi0 \
                             ,const int* const i ## a ## hi1 \
                             ,const int* const i ## a ## hi2 ) \
                             ,const int* const n ## a ## comp
#define CHFp_CONST_FRA( a ) const Real *const a \
                     ,D_DECL( const int* const i ## a ## lo0 \
                             ,const int* const i ## a ## lo1 \
                             ,const int* const i ## a ## lo2 ) \
                     ,D_DECL( const int* const i ## a ## hi0 \
                             ,const int* const i ## a ## hi1 \
                             ,const int* const i ## a ## hi2 ) \
                             ,const int* const n ## a ## comp
#define CHFp_FIA1( a ) int *const a \
                     ,D_DECL( const int* const i ## a ## lo0 \
                             ,const int* const i ## a ## lo1 \
                             ,const int* const i ## a ## lo2 ) \
                     ,D_DECL( const int* const i ## a ## hi0 \
                             ,const int* const i ## a ## hi1 \
                             ,const int* const i ## a ## hi2 )
#define CHFp_FRA1( a ) Real *const a \
                     ,D_DECL( const int* const i ## a ## lo0 \
                             ,const int* const i ## a ## lo1 \
                             ,const int* const i ## a ## lo2 ) \
                     ,D_DECL( const int* const i ## a ## hi0 \
                             ,const int* const i ## a ## hi1 \
                             ,const int* const i ## a ## hi2 )
#define CHFp_CONST_FIA1( a ) const int *const a \
                     ,D_DECL( const int* const i ## a ## lo0 \
                             ,const int* const i ## a ## lo1 \
                             ,const int* const i ## a ## lo2 ) \
                     ,D_DECL( const int* const i ## a ## hi0 \
                             ,const int* const i ## a ## hi1 \
                             ,const int* const i ## a ## hi2 )
#define CHFp_CONST_FRA1( a ) const Real *const a \
                     ,D_DECL( const int* const i ## a ## lo0 \
                             ,const int* const i ## a ## lo1 \
                             ,const int* const i ## a ## lo2 ) \
                     ,D_DECL( const int* const i ## a ## hi0 \
                             ,const int* const i ## a ## hi1 \
                             ,const int* const i ## a ## hi2 )

#endif
