/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Tues, Oct 5 1999

#include <cmath>
#include "Misc.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "BoxIterator.H"
#include "LayoutIterator.H"

/// Global variables for test output:

static const char *pgmname = "copyTest" ;
static const char *indent = "   " ,*indent2 = "      " ; 
static bool verbose = true ;


/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;

/**
   copyTest returns:
    0: all tests passed
 */
extern int
copyTest(void);


/// Code:
int
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  bool passed = true;
  int icode = 0;

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  icode = copyTest();
  if(icode != 0)
    {
      cout << indent << pgmname
           << " failed with error code " << icode
           << endl;
      passed = false;
    }

  if(passed)
    {
      cout << indent << pgmname
           << " passed all tests"
           << endl;
    }
#ifdef MPI
  MPI_Finalize();
#endif
  return 0;
}

int
copyTest()
{
  DisjointBoxLayout loLayoutFine, loLayoutCoar;
#if(CH_SPACEDIM == 2)
  Box b1(IntVect(7,6), IntVect(7,9));
  Box b2(IntVect(3,0), IntVect(3,3));
  Box b3(IntVect(9,4), IntVect(9,5));
  Box b4(IntVect(-1,4), IntVect(-1,12));
#elif(CH_SPACEDIM == 3)
  Box b1(IntVect(7,6,0), IntVect(7,9,0));
  Box b2(IntVect(3,0,0), IntVect(3,3,0));
  Box b3(IntVect(9,4,0), IntVect(9,5,0));
  Box b4(IntVect(-1,4,0), IntVect(-1,12,0));
#else
  bogus spacedim;         
#endif
  loLayoutCoar.addBox(b1,0);
  loLayoutCoar.addBox(b2,0);
  loLayoutCoar.addBox(b3,0);

  loLayoutFine.addBox(b1,0);
  loLayoutFine.addBox(b2,0);
  loLayoutFine.addBox(b3,0);
  loLayoutFine.addBox(b4,0);

  loLayoutCoar.close();
  loLayoutFine.close();

  LevelData<FArrayBox> loFabCoar(loLayoutCoar,1);
  LevelData<FArrayBox> loFabFine(loLayoutFine,1);

  DataIterator ditCoar =loFabCoar.dataIterator();

  Real fluxval = 4.77;
  int icode = 0;
  for(ditCoar.reset(); ditCoar.ok(); ++ditCoar)
    {
      FArrayBox& fab = loFabCoar[ditCoar];
      Real val;
      if(loLayoutCoar.get(ditCoar) == b2)
        val = -fluxval;
      else
        val = 0.0;

      fab.setVal(val);
      val = 1.0;
    }  

  DataIterator ditFine =loFabFine.dataIterator();
  for(ditFine.reset(); ditFine.ok(); ++ditFine)
    {
      FArrayBox& fab = loFabFine[ditFine()];
      Real val = -fluxval/2.0;
      fab.setVal(val);
    }
  loFabCoar.copyTo( loFabCoar.interval(),
                      loFabFine,
                      loFabFine.interval());

  Real eps = 1.0e-10;
  for(ditFine.reset(); ditFine.ok(); ++ditFine)
    {
      FArrayBox& fab = loFabFine[ditFine()];
      Real rmax = fab.max(0);
      Real rmin = fab.min(0);
      Real valcomp;
      if(loLayoutFine.get(ditFine()) == b2)
        valcomp = -fluxval;
      else if(loLayoutFine.get(ditFine()) == b4)
        valcomp = -fluxval/2.0;
      else
        valcomp = 0.0;

      if(Abs(rmax - valcomp) > eps) 
        {
          if(verbose)
            cout << indent2 << "max is wrong" << endl;
          icode =-1;
        }
      if(Abs(rmin - valcomp) > eps) 
        {
          if(verbose)
            cout << indent2 << "min is wrong" << endl;
          icode =-2;
        }
    }
  if(verbose)
    {
      cout << indent2 << "LoFabCoar:" << endl;
      for(ditCoar.reset(); ditCoar.ok(); ++ditCoar)
        {
          FArrayBox& fab = loFabCoar[ditCoar()];
          BoxIterator bit(fab.box());
          for(bit.reset(); bit.ok(); ++bit)
            {
              cout << indent2 << bit() << indent2 << fab(bit(),0) << endl;
            }
        }
      cout << indent2 << "LoFabFine:" << endl;
      for(ditFine.reset(); ditFine.ok(); ++ditFine)
        {
          FArrayBox& fab = loFabFine[ditFine()];
          BoxIterator bit(fab.box());
          for(bit.reset(); bit.ok(); ++bit)
            {
              cout << indent2 << bit() << indent2 << fab(bit(),0) << endl;
            }
        }
    }
  return(icode);
} 


///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
