/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::cout ;
using std::endl ;
#include "RealVect.H"
#ifdef MPI
#include <mpi.h>
#endif

/// Prototypes:

void
parseTestOptions( int argc ,char* argv[] ) ;

int
testRealVect();


/// Global variables for handling output:
static const char *pgmname = "testRealVect" ;
static const char *indent = "   ", *indent2 = "      " ;
static bool verbose = true ;


/// Code:

int
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  parseTestOptions( argc ,argv ) ;

  if( verbose )
    cout << indent2 << "Beginning " << pgmname << " ..." << endl ;

  ///
  // Run the tests
  ///
  int ret = testRealVect() ;

  if( ret == 0 )
    {
      cout << indent << pgmname << " passed all tests" << endl ;
    }
  else
    {
      cout << indent << pgmname << " failed " << ret << " test(s)" << endl ;
    }

#ifdef MPI
  MPI_Finalize();
#endif
  return ret;
}


#ifdef CH_USE_FLOAT
const Real eps = 0.75e-6 ;
#else
const Real eps = 1e-14 ;
#endif

int
testRealVect()
{
  int errors = 0 ;

  { // operator RealVect / Real
    bool err = false ;
    RealVect testrv( D_DECL( 10.0 ,20.0 ,30.0 ) );
    RealVect testdiv = testrv / 5.0 ;
    for (int dir=0; dir<SpaceDim; dir++){
      Real relerror = Abs ( 1.0 - ( testdiv[dir] / ( 2.0 * (Real)( dir+1 ) ) ) ) ;
      err |= ( relerror > eps ) ;
    }
    if( err ){
      cout << indent << pgmname << ": RealVect failed the operator/(RV,R) test" << endl;
      cout << indent2 << "info: expected (" D_TERM( << 2.0 ,<< 4.0 ,<< 6.0 ) << "), got " << testdiv << endl;
      ++errors ;
    }else{
      if( verbose ){
        cout << indent << pgmname << ": RealVect passed the operator/(RV,R) test" << endl;
      }
    }
  }

  { // operator Real / RealVect
    bool err = false ;
    RealVect testrv( D_DECL( 10.0 ,20.0 ,30.0 ) );
    RealVect testdiv = 5.0 / testrv ;
    for (int dir=0; dir<SpaceDim; dir++){
      Real relerror = Abs ( 1.0 - ( testdiv[dir] * 2.0 * (Real)( dir+1 ) ) ) ;
      err |= ( relerror > eps ) ;
    }
    if( err ){
      cout << indent << pgmname << ": RealVect failed the operator/(R,RV) test" << endl;
      cout << indent2 << "info: expected (" D_TERM( << 0.5 ,<< 0.25 ,<< 1.0/6.0 ) << "), got " << testdiv << endl;
      ++errors ;
    }else{
      if( verbose ){
        cout << indent << pgmname << ": RealVect passed the operator/(R,RV) test" << endl;
      }
    }
  }

  { // operator RealVect / RealVect
    bool err = false ;
    RealVect testrv1( D_DECL( 10.0 ,20.0 ,30.0 ) );
    RealVect testrv2( D_DECL(  5.0 , 5.0 , 5.0 ) );
    const RealVect testdiv = testrv1 / testrv2 ;
    for (int dir=0; dir<SpaceDim; dir++){
      Real relerror = Abs ( 1.0 - ( testdiv[dir] / ( 2.0 * (Real)( dir+1 ) ) ) ) ;
      err |= ( relerror > eps ) ;
    }
    if( err ){
      cout << indent << pgmname << ": RealVect failed the operator/(RV,RV) test" << endl;
      cout << indent2 << "info: expected (" D_TERM( << 2.0 ,<< 4.0 ,<< 6.0 ) << "), got " << testdiv << endl;
      ++errors ;
    }else{
      if( verbose ){
        cout << indent << pgmname << ": RealVect passed the operator/(RV,RV) test" << endl;
      }
    }
  }

  return errors ;
}


///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] )
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}
